/*-------------------------------------------------------------------
  All code is provided for EDUCATIONAL purposes ONLY. There is no
  warranty and no support.
  
  Released under the MIT License  - see
  http://www.opensource.org/licenses/mit-license.php

  thermostat4 by Jack Botner

  humidity.c - humidity related routines

  The HS1101 humidity sensor is essentially a variable capacitor
  that controls the frequency of a TLC555 timer oscillator. The
  circuit and conversion specifications come straight from the 
  HS1101 spec sheet. The oscillator clocks timer1. This is sampled
  at 1 second intervals in the timer0 interrupt handler routine.
  Once every 5 seconds the frequency sample is converted to the
  humidity reading by function calculate_humidity(). Function 
  display_humidity() is called when its time to update the display.
  
-------------------------------------------------------------------*/
#include <avr/io.h>
#include <stdlib.h>
#include <string.h>
#include "main.h"
#include "lcd.h"

#define MIN_RH					0
#define MAX_RH				   99

// Global variables
extern uint16_t	uiLMC555Frequency;      // TLC555 oscillator frequency

static volatile uint8_t	ucPercentHumidity;

// Internal functions
static uint16_t freq2rh( uint16_t uiFreq );

/*-----------------------------------------------------------------
  calculate_humidity()

  Convert the frequency from timer1 to relative humidity.
  
   Input: uiLMC555Frequency
  Output: ucPercentHumidity
-----------------------------------------------------------------*/
void calculate_humidity()
{
  ucPercentHumidity = (uint8_t) freq2rh( uiLMC555Frequency );
}

/*-----------------------------------------------------------------
  display_humidity(...)

  Format the percent humidity and display on LCD.
  
  Input: ucPercentHumidity
-----------------------------------------------------------------*/
void display_humidity()
{
  uint16_t	 	uiLen;
  char			cBuffer[8];

  lcd_gotoxy( 11, 1 );
  utoa( (unsigned) ucPercentHumidity, cBuffer, 10 );

  uiLen = strlen( cBuffer );
  if ( !uiLen || uiLen > 2 )
    return;				// what the heck?

  if ( uiLen == 1 )
  {
	// output a fixed length to avoid crap on the display
	cBuffer[1] = cBuffer[0];
	cBuffer[0] = ' ';
	cBuffer[2] = '\0';  
  }
  
  lcd_puts( cBuffer );
  lcd_putc( '%' );
}

/*---------------------------------------------------------------------
  freq2rh(...)

  Converts the frequency to relative humidity, using linear interpolation.

  Linear interpolation takes two data points, say (Xa,Ya) and (Xb,Yb),
  and the interpolant (X,Y) is given by:

           (X - Xa) (Yb - Ya)
  Y = Ya + ------------------
               (Xb - Xa)

  In this case X is frequency and Y is relative humidity.

  Input: frequency 7351..6033 Hz
---------------------------------------------------------------------*/
static uint16_t freq2rh( uint16_t uiFreq )
{
  static int16_t sFreq[11] =
  {
    7351, 7224, 7100, 6976, 6853, 6728, 6600, 6468, 6330, 6186, 6033
  };
  static int16_t sRh[11] =
  {
    MIN_RH, 10, 20, 30, 40, 50, 60, 70, 80, 90, MAX_RH
  };

  uint8_t	ii;
  uint16_t	usRH;
  int16_t	sTerm1, sTerm2, sTerm3;

  // Check for boundary values
  if ( uiFreq <= (uint16_t) sFreq[10] ) return MAX_RH;
  if ( uiFreq >= (uint16_t) sFreq[0] )  return MIN_RH;

  for ( ii=0; ii<11; ++ii )
  {
    if ( uiFreq >= (uint16_t) sFreq[ii] )
      break;
  }

  if ( uiFreq == (uint16_t) sFreq[ii] )
    return (uint16_t) sRh[ii];          // no interpolation necessary

  sTerm1 = (int16_t) uiFreq - sFreq[ii-1];
  sTerm2 = sRh[ii] - sRh[ii-1];
  sTerm3 = sFreq[ii] - sFreq[ii-1];
  usRH = (uint16_t) ( sRh[ii-1] + ( ( sTerm1 * sTerm2 ) / sTerm3 ) );

  return usRH;
}
