/*-----------------------------------------------------------------
  All code is provided for EDUCATIONAL purposes ONLY. There is no
  warranty and no support.
  
  Released under the MIT License  - see
  http://www.opensource.org/licenses/mit-license.php

  thermostat2 by Jack Botner

  interrupt.c

  This module contains all the interrupt handlers.
-----------------------------------------------------------------*/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <string.h>
#include "main.h"

// Global variables

extern MY_EEPROM_DATA stEepromVars;		// variables kept in EEPROM

extern uint8_t  ucEventFlags1, ucEventFlags2;	// system event flags
extern uint8_t  ucDelaySec;				// system response delay to temp. change
extern uint16_t uiSetTemp10C_Stage;		// used during adjustments and for display
extern uint16_t uiSetTemp10C_Final;		// used for state analysis
extern uint8_t  ucHeatLockoutSec;		// delay in restarting furnace
extern uint8_t  ucCoolLockoutSec;		// delay in restarting A/C
extern uint8_t  ucDebouncedState;		// debounced state of the switches
extern uint16_t uiAdc;					// last ADC sample
extern uint16_t uiMsPressed;			// how long encoder button pressed
extern uint16_t uiLMC555Frequency;      // Timer1 count every second

static uint8_t  ucPreviousState=0; 		// previous switch state
static uint8_t  ucNrUps=0, ucNrDowns=0; // CTS Encoder logic

// Internal Functions

static void analyze_switch(void);

/*-----------------------------------------------------------------
  ISR(TIMER0_OVF_vect)

  Timer 0 overflow interrupt handler 

  Used to provide longer intervals to drive the various events.

  System clock frequency = 8 mHz or 125ns
  Prescaler = 1024 -> timer count period = 128 us
  Timer0 counts = 196 -> interrupt period = 25.088 ms
  
  Since timer1 is needed to act as a counter for the humidity
  device, it isn't available to give us convenient 1 second
  timings. The best we can do with timer0 is 32.7ms. Here I set
  it up to produce 25.088ms interrupts. Counting to 40 gives
  us 1.00352 seconds, which is close enough. Adjusting OSCCAL
  can get us very close to 25.0ms.
-----------------------------------------------------------------*/
ISR(TIMER0_OVF_vect)
{
  static uint8_t ucTimer0Counter=0;

  // Toggle the diagnostic port
  if ( test_bit( DIAG_PORT_IN, DIAG_PIN ) )
    clear_bit( DIAG_PORT_OUT, DIAG_PIN );
  else
    set_bit( DIAG_PORT_OUT, DIAG_PIN );
	
  ++ucTimer0Counter;
  if ( ucTimer0Counter == 41 )
  {
	uiLMC555Frequency = TCNT1;        // grab the timer1 counter
	TCNT1 = 0;                        // reset timer1
	
	set_bit( ucEventFlags2, EVENT2_ONE_SECOND );
	ucTimer0Counter = 0;  
  }

  TCNT0 = INIT_TIMER0;
}

/*-----------------------------------------------------------------
  ISR(TIMER2_OVF_vect)

  Timer 2 overflow interrupt handler - tests for button presses

  System clock frequency = 8 mHz
  System clock period = 125 ns
  Prescaler = 32
  Timer count period = 8 us
  Counts = 250
  Interrupt period = 1 ms
------------------------------------------------------------------*/
ISR(TIMER2_OVF_vect)
{
  static volatile uint8_t ucFirst=1;
  static volatile uint8_t ucIndex=0;             	// index into ucState
  static volatile uint8_t ucState[MAX_CHECKS] =  	// array containing bounce status
        	{ 0, 0, 0, 0, 0, 0 };

  uint8_t ii, ucResult, ucSwIn=0;

  // We use the technique documented in "A Guide to Debouncing" by
  // Jack Ganssle to debounce up to 8 inputs (5 in this case).

  // Read current switch states
  if ( !test_bit( ENCODER_PORT, ENCODER_A ) )
    set_bit( ucSwIn, ENC_SWITCH_A );
  if ( !test_bit( ENCODER_PORT, ENCODER_B ) )
    set_bit( ucSwIn, ENC_SWITCH_B );
  if ( !test_bit( ENCODER_PORT, ENCODER_PB ) )
    set_bit( ucSwIn, ENC_SWITCH_PB );

  ucState[ucIndex] = ucSwIn;

  ++ucIndex;
  if ( ucIndex == MAX_CHECKS )
    ucIndex = 0;

  // And all columns of the array
  for ( ii=0, ucResult=0xff; ii<(MAX_CHECKS-1); ++ii )
    ucResult = ucResult & ucState[ii];

  ucDebouncedState = ucResult;

  // The 1st time only, set the previous switch state to the
  // current switch state.
  if ( ucFirst )
  {
    ucPreviousState = ucDebouncedState;
	ucFirst = 0;
  }

  // Since we're here every 2 milliseconds, keep count
  // of how many milliseconds the encoder button is pressed

  if ( test_bit( ucDebouncedState, ENC_SWITCH_PB ) )
    uiMsPressed += 2;
  else
    uiMsPressed = 0;

  if ( ucDebouncedState != ucPreviousState )
    analyze_switch();		// convert switch change to event

  TCNT2 = INIT_TIMER2;
}

/*-----------------------------------------------------------------
  ISR( ADC_vect ) - ADC conversion complete interrupt handler

  Capture the A/D result registers into static memory uiAdc.
  Schedule an event to process the result.
----------------------------------------------------------------*/
ISR( ADC_vect )
{
  // ADCL must be read before ADCH.
  uiAdc = (uint16_t) ADCL | ( (uint16_t) ADCH << 8 );

  set_bit( ucEventFlags1, EVENT_CONVERT_TEMP );
}

/*-----------------------------------------------------------------
  analyze_switch()

  Analyzes the rotary encoder to determine if the shaft rotation
  is up or down, or if the pushbutton has been pressed.
-----------------------------------------------------------------*/
static void analyze_switch()
{
  uint8_t	ucNow, ucThen;

  if ( test_bit( ucDebouncedState, ENC_SWITCH_PB ) &&
  !test_bit( ucPreviousState, ENC_SWITCH_PB ) )
  {
	// The encoder button has been pressed
	set_bit( ucEventFlags1, EVENT_BUTTON_PRESSED );
  }

  // Convert the position of the encoder to a number 0..3
  ucNow = ucDebouncedState & 0x03;
  ucThen = ucPreviousState & 0x03;

  // Given the A output in bit 0 and B output in bit 1 positions, the
  // encoder generates the values 0, 1, 3, 2 as it is rotated clockwise.

  if ( ucNow == 0 )
  {
	if ( ucThen == 2 )
	{
	  ++ucNrUps;
	  ucNrDowns = 0;
	}
	else
	{
	  if ( ucThen == 1 )
	  {
		++ucNrDowns;
		ucNrUps = 0;
	  }
	}	  
  }
  else
  {
    if ( ucNow == 1 )
    {
	  if ( ucThen == 0 )
	  {
		++ucNrUps;
		ucNrDowns = 0;
	  }
	  else
	    if ( ucThen == 3 )
	    {
		  ++ucNrDowns;
		  ucNrUps = 0;
	    }
    }
    else
	{
      if ( ucNow == 3 )
      {
	    if ( ucThen == 1 )
	    {
		  ++ucNrUps;
		  ucNrDowns = 0;
	    }
	    else
	    if ( ucThen == 2 )
	    {
		  ++ucNrDowns;
		  ucNrUps = 0;
	    }
      }
      else	// ucNow == 2
      {
	    if ( ucThen == 3 )
        {
	      ++ucNrUps;
	      ucNrDowns = 0;
        }
        else
		{
	      if ( ucThen == 0 )
	      {
	        ++ucNrDowns;
	        ucNrUps = 0;
	      }
		}		  
      }
	}
  }  			 

  if ( ucNrUps == 4 )
  {
	set_bit( ucEventFlags1, EVENT_ENCODER_UP );
	ucNrUps = ucNrDowns = 0;
  }

  if ( ucNrDowns == 4 )
  {
	set_bit( ucEventFlags1, EVENT_ENCODER_DOWN );
	ucNrUps = ucNrDowns = 0;
  }

  ucPreviousState = ucDebouncedState;
}
