/*-----------------------------------------------------------------
  All code is provided for EDUCATIONAL purposes ONLY. There is no
  warranty and no support.
  
  Released under the MIT License  - see
  http://www.opensource.org/licenses/mit-license.php

  thermostat4 by Jack Botner

  main.c

  This program implements a thermostat with the following hardware:
  - a 16 x 2 LCD display
  - a rotary encoder with pushbutton
  - a TC1047A temperature sensor
  - HS1101 humidity sensor + TLC555 oscillator
  - three relays that control the furnace/AC/Fan

  Note: All temperatures in tenths degrees C

  Processor: ATMega88PA
      Clock: Internal, 8.0 MHz

  20140105 - Adapted from thermostat3
-----------------------------------------------------------------*/
#include <stdint.h>
#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#define F_CPU 8000000UL			// 8 MHz
#include <util/delay_basic.h>
#include <stdlib.h>
#include <string.h>
#include "main.h"
#include "lcd.h"

#define DIGITAL_FILTER_SAMPLES 25	// to reduce jitter in the display

// To improve the accuracy of the A/D, measure Vref at pin 32 (AREF).
// This varies from one cpu to another. 
#define VREF_MV				1100L	// nominal 1.1V Vref, mV

#define COOL_LOCKOUT_SECS 180		// to prevent A/C from starting too soon
#define HEAT_LOCKOUT_SECS 120		// to prevent furnace from starting too soon

// System states
#define STATE_RUN			1
#define STATE_PROGRAM		2

// Global variables

MY_EEPROM_DATA stEepromVars;			// variables kept in EEPROM

uint8_t ucDebouncedState=0;    			// debounced state of the switches

static uint16_t uiActualTemp10C=220;	// current temperature, tenths degrees C
uint16_t uiMsPressed=0;					// how long encoder button pressed

uint16_t uiLMC555Frequency=6000;        // TLC555 oscillator frequency

// The following are the set temperatures in tenths degrees C. Normally
// they are identical, except when the user is turning the rotary encoder
// knob to change the temperature up or down.
// This is to prevent the user from changing the state of the
// system until the button pressing is complete. After a few seconds,
// uiSetTemp10C_Final is set from uiSetTemp10C_Stage.
uint16_t uiSetTemp10C_Stage=210;			// used during adjustments and for display
uint16_t uiSetTemp10C_Final=210;			// used for state analysis

uint8_t ucEventFlags1=0, ucEventFlags2=0;
uint16_t uiAdc=0;							// last ADC sample
static uint16_t uiSmoothedAdc=0;			// smoothed ADC value

uint8_t ucDelaySec=0;						// system response delay to temp. change
uint8_t ucHeatLockoutSec=0;					// delay in restarting furnace
uint8_t ucCoolLockoutSec=0;					// delay in restarting A/C
static uint8_t ucSystemState=STATE_RUN;		// system states

// Internal Functions

static void convert_temperature(void);
static void digital_filter_adc(void);
static void display_state(void);
static void display_temp(void);
static void state_analysis(void);
static void state_run(void);
static void state_program(void);
static void system_init(void);

/*-----------------------------------------------------------------
  main()
-----------------------------------------------------------------*/
int main( void )
{
  const char *msg1a = "Thermostat";
  const char *msg1b = "Version 4.0.1";

  uint8_t	 ii;

  system_init();          		// initialise system

  // Display the firmware version
  lcd_init( LCD_DISP_ON );
  lcd_home();
  lcd_puts( msg1a );
  lcd_gotoxy( 0, 1 );
  lcd_puts( msg1b );

  // Get the saved state variables from eeprom
  eeprom_read_block( &stEepromVars, 0, sizeof( MY_EEPROM_DATA ) );
  if ( stEepromVars.uiSetTemp10C_Heat > 400 ||		// probably EEPROM not initialized
	   stEepromVars.uiSetTemp10C_Cool > 400 ||
	   stEepromVars.ucSystemMode > 5 ||
	   stEepromVars.ucHysteresis_Heat > 20 ||
	   stEepromVars.ucHysteresis_Cool > 20 )
  {
    stEepromVars.uiSetTemp10C_Heat = 210;
    stEepromVars.uiSetTemp10C_Cool = 280;
    stEepromVars.ucSystemMode = 0;
    stEepromVars.ucHysteresis_Heat = HYSTERESIS_HEAT;
    stEepromVars.ucHysteresis_Cool = HYSTERESIS_COOL;
	stEepromVars.ucOsccal = OSCCAL_DEFAULT;
	set_bit( ucEventFlags1, EVENT_UPDATE_EEPROM );
  }

  OSCCAL = stEepromVars.ucOsccal;  // to get 25ms interrupts on timer0

  // Initialize the temperature states
  if ( test_bit( stEepromVars.ucSystemMode, MODE_COOL_ON ) )
	uiSetTemp10C_Stage = stEepromVars.uiSetTemp10C_Cool;
  else
	uiSetTemp10C_Stage = stEepromVars.uiSetTemp10C_Heat;

  uiSetTemp10C_Final = uiSetTemp10C_Stage;
  	
  // Initialize the fan relay
  if ( test_bit( stEepromVars.ucSystemMode, MODE_FAN_ON ) )
	clear_bit( RELAY_PORT, RELAY_FAN );

  // Wait a bit then clear the screen
  for ( ii=0; ii<60; ++ii )		// 2 sec delay
    _delay_loop_2( 65000 );		// 32 ms delay
  lcd_clrscr();

  // Initialize the watchdog
  wdt_enable( WDTO_500MS );
  wdt_reset();

  // Now down to business...
  while (1)
  {
    if ( ucSystemState == STATE_RUN )
	  state_run();
	else
	  state_program();
  }
}

/*-----------------------------------------------------------------
  state_run()

  The mainline when the state is "run".
-----------------------------------------------------------------*/
static void state_run()
{
  static uint8_t uc5secctr=0;
  
  wdt_reset();

  if ( test_bit( ucEventFlags1, EVENT_CONVERT_TEMP ) )
  {
	digital_filter_adc();
	convert_temperature();
	clear_bit( ucEventFlags1, EVENT_CONVERT_TEMP );
  }

  if ( test_bit( ucEventFlags1, EVENT_UPDATE_LCD ) )
  {
    display_temp();
    display_humidity();
    display_state();
	clear_bit( ucEventFlags1, EVENT_UPDATE_LCD );
  }

  if ( test_bit( ucEventFlags1, EVENT_ANALYZE_STATE ) )
  {
	state_analysis();
	clear_bit( ucEventFlags1, EVENT_ANALYZE_STATE );
  }

  if ( test_bit( ucEventFlags1, EVENT_UPDATE_EEPROM ) )
  {
	eeprom_write_block( &stEepromVars, 0, sizeof( MY_EEPROM_DATA ) );
	clear_bit( ucEventFlags1, EVENT_UPDATE_EEPROM );
  }

  if ( test_bit( ucEventFlags1, EVENT_ENCODER_UP ) )
  {
	uiSetTemp10C_Stage += SET_TEMP_DELTA;
	ucDelaySec = SYSTEM_DELAY_SEC;
    set_bit( ucEventFlags1, EVENT_UPDATE_LCD );
  	clear_bit( ucEventFlags1, EVENT_ENCODER_UP );
  }

  if ( test_bit( ucEventFlags1, EVENT_ENCODER_DOWN ) )
  {
	uiSetTemp10C_Stage -= SET_TEMP_DELTA;
	ucDelaySec = SYSTEM_DELAY_SEC;
    set_bit( ucEventFlags1, EVENT_UPDATE_LCD );
	clear_bit( ucEventFlags1, EVENT_ENCODER_DOWN );
  }

  if ( test_bit( ucEventFlags1, EVENT_BUTTON_PRESSED ) )
  {
    ucSystemState = STATE_PROGRAM;
	clear_bit( ucEventFlags1, EVENT_BUTTON_PRESSED );
  }

  if ( test_bit( ucEventFlags2, EVENT2_ONE_SECOND ) )
  {
	// Do things needed to be done every second.  
	set_bit( ucEventFlags1, EVENT_UPDATE_LCD );
	set_bit( ucEventFlags1, EVENT_ANALYZE_STATE );
	
    ++uc5secctr;
	if ( uc5secctr > 4 )
	{
	  // Schedule a 5-second event
	  set_bit( ucEventFlags2, EVENT2_5_SECONDS );
	  uc5secctr = 0;
	}
	
	// If uiSetTemp10C_Stage has changed, update uiSetTemp10C_Final after
	// the specified period.
	if ( ucDelaySec )
	{
	  -- ucDelaySec;
	  if ( !ucDelaySec )
	  {
		uiSetTemp10C_Final = uiSetTemp10C_Stage;
		if ( test_bit( stEepromVars.ucSystemMode, MODE_COOL_ON ) )
		  stEepromVars.uiSetTemp10C_Cool = uiSetTemp10C_Final;
		else
		  stEepromVars.uiSetTemp10C_Heat = uiSetTemp10C_Final;

		set_bit( ucEventFlags1, EVENT_UPDATE_EEPROM );
	  }
	}
	
	// If a lockout timer is on, decrement it each second
	if ( ucHeatLockoutSec )
	  --ucHeatLockoutSec;
	if ( ucCoolLockoutSec )
	  --ucCoolLockoutSec;

	clear_bit( ucEventFlags2, EVENT2_ONE_SECOND );
  }

  if ( test_bit( ucEventFlags2, EVENT2_5_SECONDS) )
  {
	// What we do every 5 seconds
	set_bit( ADCSRA, ADSC );	// initiate the ADC conversion
    calculate_humidity();
	
    clear_bit( ucEventFlags2, EVENT2_5_SECONDS );
  }

}

/*-----------------------------------------------------------------
  state_program()

  The mainline when the state is "program".
  
  If the button is pressed for more than 1 second it means go
  down a level in the menu and program that item. Shorter
  press means select the next item. uiMsPressed is constantly
  updated in the 2ms timer interrupt handler.
-----------------------------------------------------------------*/
static void state_program()
{
  program_mode();               // program heat/cool/off

  if ( uiMsPressed < 1000 )
    program_fan();              // program fan on/off

  if ( uiMsPressed < 1000 )
	program_hysteresis();       // program hysteresis
	
  if ( uiMsPressed < 1000 )
	program_osccal();           // program osccal

  lcd_clrscr();
  // Set the system delay, which determines how long to wait
  // before the change takes effect. This is in case the setting
  // is changed again too quickly.
  ucDelaySec = SYSTEM_DELAY_SEC;
  ucSystemState = STATE_RUN;
}

/*-----------------------------------------------------------------
  system_init()

  PORTB: PB0 (input) Rotary Encoder "A"
         PB1 (input) Rotary Encoder "B"
         PB2 (input) Rotary Encoder "S1"
         PB3 (input) ISP MOSI
         PB4 (input) ISP MISO
         PB5 (input) ISP SCK
         PB6 (input) (not used)
         PB7 (input) (not used)

  PORTC: PC0 (input) (ADC0) temperature sensor
         PC1 (output) HEAT relay
         PC2 (output) COOL relay
         PC3 (output) FAN relay
         PC4 (input) (not used)
         PC5 (input) (not used)
         PC6 (input) (ISP !reset) 

  PORTD: PD0 (output) LCD DB4
         PD1 (output) LCD DB5
         PD2 (output) LCD DB6
         PD3 (output) LCD DB7
         PD4 (output) LCD RS
         PD5 (input) (T1) humidity sensor signal
         PD6 (output) LCD RW
         PD7 (output) LCD E
-----------------------------------------------------------------*/
static void system_init()
{
  // The fuses are set to provide an internal clock, running at
  // 8.0 MHz with a prescaler of 8. Change the prescaler from
  // 8 to 1. The procedure requires two steps as per the device
  // specifications.
  CLKPR = 0b10000000;       // set clock prescaler change enable
  CLKPR = 0b00000000;       // set prescaler to 1
  
  // I like to set unused pins to input with pullup.
  DDRB  = 0b00000000;		// pb: 0=input, 1=output
  PORTB = 0b11111111;       // initialize PortB

  DDRC  = 0b00101110;		// pc: 0=input, 1=output
  PORTC = 0b01011110;       // initialize PortC

  DDRD  = 0b11011111;		// pd: 0=input, 1=output
  PORTD = 0b00000000;       // initialize PortD

  // Set up PC0/ADC0 as analog input, Vref=1.1V
  ADMUX  = 0b11000000;		// Vref=1.1V; result right adjusted; select ADC0
  ADCSRA = 0b10001110;		// Enable ADC; Enable interrupt; prescaler=64
  ADCSRB = 0b00000000;		// free running mode
  DIDR0  = 0b00000001;		// disable digital I/O on PC0

  // Initialize Timer0. 
  // Timer0 is used to create 25ms intervals for scheduling events
  TCCR0A = 0b00000000;		// normal port operation
  TCCR0B = 0b00000101;		// prescaler = 1024
  TCNT0 = INIT_TIMER0;
  set_bit( TIMSK0, TOIE0 );	// enable timer/counter0 overflow interrupt

  // Initialize Timer1.
  // Initialize Timer1. Timer1 is used as a counter for the LMC555 oscillator.
  TCCR1A = 0b00000000;		// normal operation, no OC1A, no OC1B
  TCCR1B = 0b00000110;      // external clock on T1 pin, clock on falling edge
  TCCR1C = 0;				// not comparing anything
  TCNT1 = 0;
  TIMSK1 = 0b00000000;      // no timer/counter1 overflow interrupt

  // Initialize Timer2. Timer2 is used to generate interrupts
  // every 1 millisecond for switch debouncing.
  TCCR2A = 0b00000000;		// normal port operation
  TCCR2B = 0b00000011;		// normal operation, prescaler=32
  TCNT2 = INIT_TIMER2;
  set_bit( TIMSK2, TOIE2 );	// enable timer/counter2 overflow interrupt

  sei();					// enable global interrupts
}

/*-----------------------------------------------------------------
  convert_temperature()

  Convert the temperature from the A/D output to 10ths of degrees C.
  ADC0 is monitoring TC1047A Vout signal, and Vref = 1.1V (internal).

  First we need to convert the A/D voltage to millivolts, then
  to tenths of degrees C.
  
  The formula for voltage to temperature conversion is 
  
				  V - 500
			T = ------------
				   10.0
				   
  where T is temperature in degrees C, and V is Vout in millivolts.

  Input: uiSmoothedAdc from the adc filter.

  Output: uiActualTemp10C
-----------------------------------------------------------------*/
static void convert_temperature()
{
  uint32_t ulVoutMv;
  uint16_t uiTemp10C;

  // Input is in the range 0..1023 which represents a voltage of 0..1.10
  // volts (Vref). Scale the adc value to a voltage value (millivolts).

  ulVoutMv = ( (uint32_t) uiSmoothedAdc * VREF_MV ) / 1024L;

  // Derive temperature (tens of degrees C) from the formula above.
  // Multiply numerator and denominator by 100 to handle decimals.
  // Multiply numerator by 10 to get tens of degrees.
  
  //uiTempTensC = (unsigned int) ( ( ( ulVoutMv - 500L ) * 1000L ) / 1000L );
  uiTemp10C = (uint16_t) ( ulVoutMv - 500L );

  if ( uiTemp10C > 999 )
    uiTemp10C = 999;		// sensor may be disconnected

  uiActualTemp10C = uiTemp10C - TEMP_CORRECTION;
}

/*-----------------------------------------------------------------
  display_temp()

  Format the tenths of degrees and display on LCD.
-----------------------------------------------------------------*/
static void display_temp()
{
  const char 	*msg2a = "Tmp        ";
  const char 	*msg2b = "Set";
  char		 	buffer[16];
  uint16_t	 	uiLen;

  // Display the actual temperature on line 1

  lcd_gotoxy( 0, 0 );
  lcd_puts( msg2a );

  utoa( uiActualTemp10C, buffer, 10 );

  // Insert a decimal point in the result
  uiLen = strlen( buffer );
  if ( uiLen )
    uiLen -= 1;
  buffer[uiLen+1] = buffer[uiLen];
  buffer[uiLen] = '.';
  buffer[uiLen+2] = '\0';

  lcd_gotoxy( 4, 0 );
  lcd_puts( buffer );

  // Display the set temperature on line 2

  lcd_gotoxy( 0, 1 );
  lcd_puts( msg2b );

  utoa( uiSetTemp10C_Stage, buffer, 10 );

  // Insert a decimal point in the result
  uiLen = strlen( buffer );
  if ( uiLen )
    uiLen -= 1;
  buffer[uiLen+1] = buffer[uiLen];
  buffer[uiLen] = '.';
  buffer[uiLen+2] = '\0';

  lcd_gotoxy( 4, 1 );
  lcd_puts( buffer );
}

/*-----------------------------------------------------------------
  display_state()

  Display the system mode on the LCD.
  In the top right part of the display.
-----------------------------------------------------------------*/
static void display_state()
{
  const char *msg3a = "Heat";
  const char *msg3b = "Cool";
  const char *msg3c = " Off";
  const char *msg3d = "On";
  const char *msg3e = "  ";

  lcd_gotoxy( 9, 0 );
  if ( test_bit( stEepromVars.ucSystemMode, MODE_HEAT_ON ) )
    lcd_puts( msg3a );
  else
    if ( test_bit( stEepromVars.ucSystemMode, MODE_COOL_ON ) )
      lcd_puts( msg3b );
    else
      lcd_puts( msg3c );

  lcd_gotoxy( 14, 0 );
  if ( test_bit( RELAY_PORT, RELAY_HEAT ) && test_bit( RELAY_PORT, RELAY_COOL ) )
    lcd_puts( msg3e );
  else
    lcd_puts( msg3d );
}

/*-----------------------------------------------------------------
  state_analysis()

  Decides if the cool/heat relays should be on or off.
-----------------------------------------------------------------*/
static void state_analysis()
{
  if ( test_bit( stEepromVars.ucSystemMode, MODE_HEAT_ON ) )
  {
	// mode=Heat
    if ( test_bit( RELAY_PORT, RELAY_HEAT ) )
	{
	  // The heat relay is off. If the temperature drops 1/2 degree C
	  // below our setting, turn it on.
	  if ( uiActualTemp10C < ( uiSetTemp10C_Final - stEepromVars.ucHysteresis_Heat ) )
	  {
	    if ( !ucHeatLockoutSec )				// if not locked out
		  clear_bit( RELAY_PORT, RELAY_HEAT );	// turn heat on
	  }
	}
	else
	{
	  // The heat relay is on. If the temperature is more than specified
	  // degrees C above our setting, turn it off.
	  if ( uiActualTemp10C > ( uiSetTemp10C_Final + stEepromVars.ucHysteresis_Heat ) )
	  {
		set_bit( RELAY_PORT, RELAY_HEAT );		// turn heat off
		ucHeatLockoutSec = HEAT_LOCKOUT_SECS;	// keep it off for awhile	
	  }
	}
  }
  else
  {
    // !mode=Heat
    if ( !test_bit( RELAY_PORT, RELAY_HEAT ) )	// the heat relay is on
	{
	  set_bit( RELAY_PORT, RELAY_HEAT );		// turn it off
	  ucHeatLockoutSec = HEAT_LOCKOUT_SECS;		// keep it off for awhile	
	}
  }

  if ( test_bit( stEepromVars.ucSystemMode, MODE_COOL_ON ) )
  {
	// The cool switch is on
    if ( test_bit( RELAY_PORT, RELAY_COOL ) )
	{
	  // The cool relay is off. If the temperature rises specified degrees C
	  // above our setting, turn it on.
	  if ( uiActualTemp10C > ( uiSetTemp10C_Final + stEepromVars.ucHysteresis_Cool ) )
	  {
	    if ( !ucCoolLockoutSec )				// if not locked out
		  clear_bit( RELAY_PORT, RELAY_COOL );	// turn cool on
	  }
	}
	else
	{
	  // The cool relay is on. If the temperature drops more than 1/2
	  // degree C below our setting, turn it off.
	  if ( uiActualTemp10C < ( uiSetTemp10C_Final - stEepromVars.ucHysteresis_Cool ) )
	  {
		set_bit( RELAY_PORT, RELAY_COOL );		// turn cool off
		ucCoolLockoutSec = COOL_LOCKOUT_SECS;	// keep it off for awhile	
      }
	}
  }
  else
  {
    // The cool switch is off
    if ( !test_bit( RELAY_PORT, RELAY_COOL ) )	// the cool relay is on
	{
	  set_bit( RELAY_PORT, RELAY_COOL );		// turn it off
	  ucCoolLockoutSec = COOL_LOCKOUT_SECS;		// keep it off for awhile	
	}
  }
}

/*---------------------------------------------------------------------
  digital_filter_adc()

  Why do we need this? Because the output of the adc jitters one or
  two counts each sample, which translates to one or two tenths of a
  degree C on the display. It goes up and down apparently randomly,
  which is very annoying. According to the Atmel spec, the adc has a
  number of error modes which result in the lsb being uncertain.
---------------------------------------------------------------------*/
static void digital_filter_adc()
{
  static uint16_t uiTable[DIGITAL_FILTER_SAMPLES];
  static uint8_t  ucIndex=0, ucEntries=0;
  uint8_t		  ii;
  uint32_t		  ulSum;

  // Add the entry to the table, wrap-around
  uiTable[ucIndex] = uiAdc;
  ++ucIndex;
  if ( ucIndex >= DIGITAL_FILTER_SAMPLES )
    ucIndex = 0;
  if ( ucEntries < DIGITAL_FILTER_SAMPLES )
    ++ucEntries;

  // Average the table
  ulSum = 0;
  for ( ii=0; ii<ucEntries; ++ii )
    ulSum += uiTable[ii];

  uiSmoothedAdc = (unsigned int) ( ulSum / ucEntries );
}
